﻿// Copyright (c) 2010, Chris Sells, Phani Raj Yayavaram Narasimha.
// You may use this code as you see fit. Enjoy.
// Inspired by a series of posts by Joe Cheng about integrating Windows Live Writer and AtomPub:
// http://jcheng.wordpress.com/2007/10/15/how-wlw-speaks-atompub-introduction/

using System;
using System.Collections.Generic;
using System.Data.Objects;
using System.Data.Services;
using System.Linq;
using System.Reflection;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Description;
using System.ServiceModel.Dispatcher;
using System.Text;
using System.Xml;
using System.Xml.Linq;

namespace ODataBlogging {
  public class ODataBloggingMessageInspector : IDispatchMessageInspector {
    static readonly string nsAtom = "http://www.w3.org/2005/Atom";
    static readonly string nsMetadata = "http://schemas.microsoft.com/ado/2007/08/dataservices/metadata";
    static readonly string nsDataServices = "http://schemas.microsoft.com/ado/2007/08/dataservices";

    Type containerType;

    public ODataBloggingMessageInspector(Type containerType = null) {
      this.containerType = containerType;
    }

    static bool IsAtomProperty(string propertyName) {
      switch (propertyName.ToLower()) {
        case "updated":
        case "content":
        case "title":
        case "published":
        case "summary":
          // TODO: the rest, including nested properties inside <author />
          return true;
      }

      return false;
    }

    static IEnumerable<PropertyInfo> GetObjectSetProperties(Type type) {
      // NOTE: the specific class passed to ObjectSet<T> isn't important -- just that it's a class
      Type objectSetGenericTypeDefinition = typeof(ObjectSet<Type>).GetGenericTypeDefinition();
      return from p in type.GetProperties()
             where p.PropertyType.IsGenericType &&
                 p.PropertyType.GetGenericTypeDefinition() == objectSetGenericTypeDefinition
             select p;
    }

    IEnumerable<EntityAtomMappingAttribute> GetAtomMappingsForEntitySet(InstanceContext instanceContext, string entitySetName) {
      // Return the manual property mappings
      Type dataServiceType = ((DataServiceHost)instanceContext.Host).Description.ServiceType;
      var attribs = dataServiceType.GetCustomAttributes(typeof(EntityAtomMappingAttribute), true).
          Cast<EntityAtomMappingAttribute>().Where(attrib => attrib.EntitySetName == entitySetName);
      foreach (var attrib in attribs) {
        yield return attrib;
      }

      if (containerType == null) { yield break; }

      // Return the atomatic property mappings
      PropertyInfo objectSetProperty = GetObjectSetProperties(containerType).FirstOrDefault(p => string.Compare(p.Name, entitySetName, true) == 0);
      if (objectSetProperty == null) { yield break; }
      Type entityType = objectSetProperty.PropertyType.GetGenericArguments()[0];
      foreach (var prop in entityType.GetProperties()) {
        if (IsAtomProperty(prop.Name)) {
          // NOTE: we're returning potentially duplicate entries here, but we're returning
          // the manually mapped ones first, so they should be found first and the dupes shouldn't matter
          yield return new EntityAtomMappingAttribute(entitySetName, prop.Name, prop.Name.ToLower());
        }
      }
    }

    class Writer : BodyWriter {
      string content;

      public Writer(string content)
        : base(false) {
        this.content = content;
      }

      protected override void OnWriteBodyContents(XmlDictionaryWriter writer) {
        writer.WriteStartElement("Binary");
        byte[] buffer = Encoding.UTF8.GetBytes(content);
        writer.WriteBase64(buffer, 0, buffer.Length);
        writer.WriteEndElement();
      }
    }

    #region IDispatchMessageInspector Members

    public object AfterReceiveRequest(ref Message request, IClientChannel channel, InstanceContext instanceContext) {
      string entitySetName = request.Properties.Via.Segments.Last();
      if (entitySetName.Contains('(')) {
        entitySetName = entitySetName.Substring(0, entitySetName.IndexOf('('));
      }
      else if (entitySetName.EndsWith("/")) {
        entitySetName = entitySetName.Substring(0, entitySetName.Length - 1);
      }

      // check for mappings
      EntityAtomMappingAttribute[] atomMappings = GetAtomMappingsForEntitySet(instanceContext, entitySetName).ToArray();
      HttpRequestMessageProperty httpmsg = (HttpRequestMessageProperty)request.Properties[HttpRequestMessageProperty.Name];

      // we're not supporting optimistic concurrency, so drop "If-Match"
      var header = httpmsg.Headers.AllKeys.FirstOrDefault(n => string.Compare(n, "If-Match") == 0);
      if (header != null) { httpmsg.Headers.Remove(header); }

      if (atomMappings.Length != 0 && String.Compare(httpmsg.Method, "GET", true) != 0 && !request.IsEmpty) {
        // grab the data
        XmlDictionaryReader reader = request.GetReaderAtBodyContents();
        reader.ReadStartElement();
        string content = Encoding.UTF8.GetString(reader.ReadContentAsBase64());
        reader.Close();

        // migrate the data into properties
        XElement entryElement = XElement.Parse(content);
        XElement propertiesElement = new XElement(XName.Get("properties", nsMetadata));

        foreach (var atomMapping in atomMappings) {
          XElement element = entryElement.Element(XName.Get(atomMapping.AtomPropertyName, nsAtom));
          if (element == null) { continue; }
          string value = element.Value;
          XName xn = XName.Get(atomMapping.EntityPropertyName, nsDataServices);
          propertiesElement.Add(new XElement(xn, value));
        }

        XElement contentElement = entryElement.Element(XName.Get("content", nsAtom));
        XAttribute contentTypeAttribute = contentElement.Attribute(XName.Get("type"));
        contentTypeAttribute.Value = "application/xml";
        string actualHTML = contentElement.Value;
        contentElement.Value = "";
        contentElement.Add(propertiesElement);
        string entityXml = entryElement.ToString();

        // create a new message with the updated data
        request = Message.CreateMessage(MessageVersion.Default, "", new Writer(entityXml));
        request.Properties[WebBodyFormatMessageProperty.Name] = new WebBodyFormatMessageProperty(WebContentFormat.Raw);
        HttpRequestMessageProperty httpRequest = new HttpRequestMessageProperty();
        httpRequest.Headers[System.Net.HttpRequestHeader.ContentType] = "application/atom+xml";
        request.Properties.Add(HttpRequestMessageProperty.Name, httpRequest);
      }

      if (String.Compare(httpmsg.Method, "GET", true) == 0) {
        // Order by the "published" date in descending order as the Atom convention proscribes
        var publishedMapping = atomMappings.FirstOrDefault(m => string.Compare(m.AtomPropertyName, "published", true) == 0);
        string publishedEntityFieldName = publishedMapping == null ? null : publishedMapping.EntityPropertyName;

        if (publishedEntityFieldName != null) {
          request.Properties.Via = new Uri(request.Properties.Via.OriginalString + "?$orderby=" + publishedEntityFieldName + " desc");
        }
      }

      return atomMappings;
    }

    public void BeforeSendReply(ref Message reply, object correlationState) {
      EntityAtomMappingAttribute[] atomMappings = (correlationState ?? new EntityAtomMappingAttribute[0]) as EntityAtomMappingAttribute[];
      if (atomMappings.Length != 0) {
        XmlDictionaryReader reader = reply.GetReaderAtBodyContents();
        reader.ReadStartElement();
        string content = Encoding.UTF8.GetString(reader.ReadContentAsBase64());
        reader.Close();

        XElement atomDocument = XElement.Parse(content);
        XName xnEntry = XName.Get("entry", nsAtom);
        XName xnProperties = XName.Get("properties", nsMetadata);

        if (atomDocument.Name == xnEntry) {
          ReWriteEntryElements(atomMappings, xnProperties, atomDocument);
        }
        else {
          foreach (XElement entryElement in atomDocument.Elements(xnEntry)) {
            ReWriteEntryElements(atomMappings, xnProperties, entryElement);
          }
        }

        content = atomDocument.ToString();
        Message newreply = Message.CreateMessage(MessageVersion.None, "", new Writer(content));
        newreply.Properties.CopyProperties(reply.Properties);
        reply = newreply;
      }
    }

    static void ReWriteEntryElements(EntityAtomMappingAttribute[] atomMappings, XName xnProperties, XElement entryElement) {
      XElement contentElement = entryElement.Element(XName.Get("content", nsAtom));
      XElement propertiesElement = contentElement.Element(xnProperties);

      propertiesElement.Remove(); // So we can read the contents but reset the content value itself
      XAttribute contentTypeAttribute = contentElement.Attribute(XName.Get("type"));
      contentTypeAttribute.Value = "html"; // TODO: parameterizable?
      contentElement.Value = "";

      foreach (var atomMapping in atomMappings) {
        XElement propertyElement = propertiesElement.Element(XName.Get(atomMapping.EntityPropertyName, nsDataServices));
        if (propertyElement == null) { continue; }

        XName xnAtomProperty = XName.Get(atomMapping.AtomPropertyName, nsAtom);
        XElement atomElement = entryElement.Element(xnAtomProperty);
        if (atomElement == null) {
          if (!string.IsNullOrWhiteSpace(propertyElement.Value)) {
            entryElement.Add(new XElement(xnAtomProperty, GetAtomFormattedValue(xnAtomProperty, propertyElement.Value)));
          }
        }
        else {
          if (!string.IsNullOrWhiteSpace(propertyElement.Value)) {
            atomElement.Value = GetAtomFormattedValue(xnAtomProperty, propertyElement.Value);
          }
          else {
            // TODO: wrong to do this on required ATOM properties?
            atomElement.Remove();
          }
        }
      }
    }

    static string GetAtomFormattedValue(XName xn, string s) {
      if (xn == XName.Get("published", nsAtom) || xn == XName.Get("updated", nsAtom)) {
        return XmlConvert.ToString(DateTime.Parse(s), XmlDateTimeSerializationMode.Local);
      }
      else { return s; }
    }

    #endregion
  }

  [AttributeUsage(AttributeTargets.Class, AllowMultiple = true)]
  public class EntityAtomMappingAttribute : Attribute {
    public string EntitySetName { get; set; }
    public string AtomPropertyName { get; set; }
    public string EntityPropertyName { get; set; }
    public EntityAtomMappingAttribute(string entitySetName, string entityPropertyName, string atomPropertyName) {
      this.EntitySetName = entitySetName;
      this.EntityPropertyName = entityPropertyName;
      this.AtomPropertyName = atomPropertyName;
    }
  }

  [AttributeUsage(AttributeTargets.Class)]
  public class ODataBloggingServiceBehavior : Attribute, IServiceBehavior {
    Type containerType;

    public ODataBloggingServiceBehavior() : this(null) { }

    public ODataBloggingServiceBehavior(Type containerType) {
      this.containerType = containerType;
    }

    #region IServiceBehavior Members

    void IServiceBehavior.AddBindingParameters(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase, System.Collections.ObjectModel.Collection<ServiceEndpoint> endpoints, System.ServiceModel.Channels.BindingParameterCollection bindingParameters) {
    }

    void IServiceBehavior.ApplyDispatchBehavior(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase) {
      foreach (ChannelDispatcher cd in serviceHostBase.ChannelDispatchers) {
        foreach (EndpointDispatcher ed in cd.Endpoints) {
          ed.DispatchRuntime.MessageInspectors.Add(new ODataBloggingMessageInspector(containerType));
        }
      }
    }

    void IServiceBehavior.Validate(ServiceDescription serviceDescription, ServiceHostBase serviceHostBase) {
    }

    #endregion
  }
}
